local Tinkr, _Bastion = ...

---@class Bastion
local Bastion = _Bastion

---@alias BigWigsBars table<string, { spellId: number, duration: number, startTime: number, pauseTime: number | nil }>

---@class ContentUtils
local ContentUtils = {
    ---@type InterruptList
    interruptList = {},
    ---@type TankBusterList
    tankBusterList = {},
    ---@type ImportantAuraList
    importantAuras = {},
    ---@type BigWigsBars
    bars = {},
    isBigWigsEnabled = false
}

ContentUtils.__index = ContentUtils

local function BWEventCallBack(event, ...)
    if event == "BigWigs_StartBar" then
        local addon, spellId, text, duration, icon, isCD = ...
        local now = GetTime()

        print('START BAR', text)

        ContentUtils.bars[text] = ContentUtils.bars[text] or {}

        ContentUtils.bars[text].duration = duration
        ContentUtils.bars[text].startTime = now
        ContentUtils.bars[text].spellId = spellId
    elseif event == "BigWigs_StopBar" then
        local addon, text = ...

        print('STOP BAR', text)

        if ContentUtils.bars[text] then
            ContentUtils.bars[text] = nil
        end
    elseif event == "BigWigs_PauseBar" then
        local addon, text = ...

        print('PAUSE BAR', text)

        if ContentUtils.bars[text] then
            ContentUtils.bars[text].pauseTime = GetTime()
        end
    elseif event == "BigWigs_ResumeBar" then
        local addon, text = ...

        print('RESUME BAR', text)

        if ContentUtils.bars[text] and ContentUtils.bars[text].pauseTime then
            local pauseTime = ContentUtils.bars[text].pauseTime
            local startTime = ContentUtils.bars[text].startTime
            local duration = ContentUtils.bars[text].duration

            local newDuration = duration - (pauseTime - startTime)

            ContentUtils.bars[text].pauseTime = nil
            ContentUtils.bars[text].duration = newDuration
            ContentUtils.bars[text].startTime = GetTime()
        end
    elseif event == "BigWigs_StopBars" or event == "BigWigs_OnBossDisable" then
        ContentUtils.bars = {}
    end
end

---@type InterruptList
local _interruptList = Tinkr:require("scripts/bastion/ContentUtils/lists/interrupt-list", Bastion)
---@type TankBusterList
local _tankBusterList = Tinkr:require("scripts/bastion/ContentUtils/lists/tank-buster-list", Bastion)
---@type ImportantAuraList
local _importantAuraList = Tinkr:require("scripts/bastion/ContentUtils/lists/important-aura-list", Bastion)


---@return ContentUtils
function ContentUtils:New()
    local self = setmetatable({}, ContentUtils)

    self.interruptList = _interruptList
    self.tankBusterList = _tankBusterList
    self.importantAuras = _importantAuraList
    if BigWigsLoader then
        print("BigWigs Enabled!")
        self.isBigWigsEnabled = true
        BigWigsLoader.RegisterMessage({}, 'BigWigs_StartBar', BWEventCallBack)
        BigWigsLoader.RegisterMessage({}, 'BigWigs_StopBar', BWEventCallBack)
        BigWigsLoader.RegisterMessage({}, 'BigWigs_PauseBar', BWEventCallBack)
        BigWigsLoader.RegisterMessage({}, 'BigWigs_ResumeBar', BWEventCallBack)
        BigWigsLoader.RegisterMessage({}, 'BigWigs_StopBars', BWEventCallBack)
        BigWigsLoader.RegisterMessage({}, 'BigWigs_OnBossDisable', BWEventCallBack)
    end

    return self
end

---@param unit Unit
---@param percent number
---@return Interrupt | nil
function ContentUtils:CastingCriticalStop(unit, percent)
    local castingSpell = unit:GetCastingOrChannelingSpell()

    local npcSpells = self.interruptList[unit:GetID()]

    if npcSpells and castingSpell then
        local spellID = castingSpell:GetID()

        local spellTraits = npcSpells[spellID]

        if not spellTraits then
            return nil
        end

        local castPercent = unit:GetChannelOrCastPercentComplete()

        if castPercent >= percent then
            return spellTraits
        end
    end

    return nil
end

function ContentUtils:RefreshBars()
    if self.isBigWigsEnabled then
        for key, value in pairs(self.bars) do
            local expireTime = value.duration + value.startTime

            if expireTime < GetTime() then
                self.bars[key] = nil
            end
        end
    end
end

---@param spellId number
---@param timeUntilCast number
---@return boolean
function ContentUtils:IsSpellCastSoon(spellId, timeUntilCast)
    if self.isBigWigsEnabled then
        self:RefreshBars()
        for key, value in pairs(self.bars) do
            if value.spellId == spellId and not value.pauseTime then
                local expireTime = value.duration + value.startTime

                local timeUntilSpell = expireTime - GetTime()
    
                if timeUntilSpell > 0 and timeUntilSpell <= timeUntilCast then
                    return true
                end
            end
        end
    end
    return false
end

---@param unit Unit
---@param timeUntilCast? number
---@return TankBuster | nil
function ContentUtils:CastingTankBuster(unit, timeUntilCast)
    local npcSpells = self.tankBusterList[unit:GetID()]

    if npcSpells then
        -- If timeUntilCast is provided, check if any of this NPC
        if timeUntilCast then
            for key, value in pairs(self.bars) do
                local spellTraits = npcSpells[value.spellId]

                if spellTraits and not value.pauseTime and ((value.startTime + value.duration) - GetTime()) <= timeUntilCast then
                    return spellTraits
                end
            end
        end

        local castingSpell = unit:GetCastingOrChannelingSpell()

        if castingSpell then
            if npcSpells[castingSpell:GetID()] then return npcSpells[castingSpell:GetID()] end
        end

    end

    return nil
end

---@param friend Unit
---@return ImportantAura | nil
function ContentUtils:HasImportantAura(friend)
    local traits = nil
    friend:GetAuras():ForEach(function(spellId)
        local auraTraits = self.importantAuras[spellId]
        if auraTraits then traits = auraTraits end
    end)
    return traits
end

return ContentUtils
